# SPDX-FileCopyrightText: 2017 Scott Shawcroft, written for Adafruit Industries
# SPDX-FileCopyrightText: Copyright (c) 2023 Phil Underwood for Underwood Underground
#
# SPDX-License-Identifier: MIT
"""
`async_button`
================================================================================

a library for reading buttons using asyncio


* Author(s): Phil Underwood

Implementation Notes
--------------------

**Software and Dependencies:**

* Adafruit CircuitPython firmware for the supported boards:
  https://circuitpython.org/downloads

* CircuitPython asyncio module:
  https://github.com/adafruit/Adafruit_CircuitPython_asyncio
"""

# imports

__version__ = "1.0.0"
__repo__ = "https://github.com/furbrain/CircuitPython_async_button.git"

import time

import asyncio

try:
    from typing import Dict, Sequence, Awaitable
except ImportError:
    pass

import digitalio
import keypad
from microcontroller import Pin
import countio


class SimpleButton:
    """
    Asynchronous interface to a button or other IO input. This does not create a background
    task.
    """

    def __init__(self, pin: Pin, value_when_pressed: bool, pull: bool = True):
        """

        :param Pin pin: Pin to wait for
        :param bool value_when_pressed: ``True`` if the pin reads high when the key is pressed.
          ``False`` if the pin reads low (is grounded) when the key is pressed.
        :param bool pull: ``True`` if an internal pull-up or pull-down should be enabled on the
          pin. A pull-up will be used if value_when_pressed is False; a pull-down will be used if
          it is ``True``. If an external pull is already provided for the pin, you can set
          pull to ``False``. However, enabling an internal pull when an external one is already
          present is not a problem; it simply uses slightly more current.
        """
        self.pin: Pin = pin
        self.value_when_pressed = value_when_pressed
        if pull:
            self.pull = digitalio.Pull.DOWN if value_when_pressed else digitalio.Pull.UP
        else:
            self.pull = None

    async def pressed(self):
        """
        Wait until button is pressed
        """
        edge = countio.Edge.RISE if self.value_when_pressed else countio.Edge.FALL
        with countio.Counter(self.pin, edge=edge, pull=self.pull) as counter:
            while True:
                if counter.count > 0:
                    return
                await asyncio.sleep(0)

    async def released(self):
        """
        Wait until button is released
        """
        edge = countio.Edge.FALL if self.value_when_pressed else countio.Edge.RISE
        with countio.Counter(self.pin, edge=edge, pull=self.pull) as counter:
            while True:
                if counter.count > 0:
                    return
                await asyncio.sleep(0)


class Button:
    # pylint: disable
    """
    This object will monitor the specified pin for changes and will report
    single, double, triple and long_clicks. It creates a background `asyncio` process
    that will monitor the button. The `events` chapter in the documentation shows when the
    various events are triggered
    """

    PRESSED = 1  #: Button has been pressed
    RELEASED = 2  #: Button has been released
    SINGLE = 4  #: Single click
    DOUBLE = 8  #: Double click
    TRIPLE = 16  #: Triple click
    LONG = 32  #: Long click
    ANY_CLICK = (
        SINGLE,
        DOUBLE,
        TRIPLE,
        LONG,
    )  #: Any of `SINGLE`, `DOUBLE`, `TRIPLE` or `LONG`
    ALL_EVENTS = (PRESSED, RELEASED, SINGLE, DOUBLE, TRIPLE, LONG)  #: Any event

    def __init__(
        self,
        pin: Pin,
        value_when_pressed: bool,
        *,
        pull: bool = True,
        double_click_max_duration=0.5,
        long_click_min_duration=2.0,
        double_click_enable: bool = True,
        triple_click_enable: bool = False,
        long_click_enable: bool = False,
    ):
        """
        Create the button object and start the background async process, this object must be
        created only when the asyncio event loop is running

        :param Pin pin: the pin to be monitored
        :param bool value_when_pressed: ``True`` if the pin reads high when the key is pressed.
          ``False`` if the pin reads low (is grounded) when the key is pressed.
        :param bool pull: ``True`` if an internal pull-up or pull-down should be enabled on
          the pin. A pull-up will be used if ``value_when_pressed`` is ``False``; a pull-down
          will be used if it is True. If an external pull is already provided for the pins,
          you can set pull to ``False``. However, enabling an internal pull when an external one
          is already present is not a problem; it simply uses slightly more current. Default is True
        :param float double_click_max_duration: how long in seconds before a second click is
          registered as a double click (this is also the value used for triple clicks.
          Default is 0.5 seconds
        :param float long_click_min_duration: how long in seconds the button must be pressed before
          a long_click is triggered. Default is 2 seconds.
        :param bool double_click_enable: Whether double clicks are detected. Default is True
        :param bool triple_click_enable: Whether triple clicks are detected. Default is False
        :param bool long_click_enable: Whether long clicks are detected. Default is False
        """
        self.pin = pin
        self.value_when_pressed = value_when_pressed
        #: Maximum separation between two clicks to register as double in seconds (default is 0.5s)
        self.double_click_max_duration = double_click_max_duration
        #: Minimum duration for a click to register as a long click in seconds. Default is 2s
        self.long_click_min_duration = long_click_min_duration
        if not double_click_enable and triple_click_enable:
            raise ValueError("Must have double click enabled to use triple click")
        self.click_enabled = {
            self.DOUBLE: double_click_enable,
            self.TRIPLE: triple_click_enable,
            self.LONG: long_click_enable,
        }
        self.keys = keypad.Keys(
            (self.pin,), value_when_pressed=value_when_pressed, pull=pull
        )
        self.monitor_task = asyncio.create_task(self._monitor())
        self.events = {
            x: asyncio.Event()
            for x in (
                self.SINGLE,
                self.DOUBLE,
                self.TRIPLE,
                self.LONG,
                self.PRESSED,
                self.RELEASED,
            )
        }
        self.last_click = self.SINGLE
        self.pressed = False

    async def _monitor(self):
        """
        This is the main background task that monitors key presses and releases
        """
        evt = keypad.Event(0, False)
        last_click_tm = -100000
        long_click_due = None
        while True:
            if self.keys.events.get_into(evt):
                if evt.pressed:
                    self._trigger(self.PRESSED)
                    now = time.monotonic()
                    # print(now, self.last_click_tm, self.double_click_max_duration)
                    if now - last_click_tm < self.double_click_max_duration:
                        self._increase_clicks()
                    else:
                        self.last_click = self.SINGLE
                    last_click_tm = now
                    long_click_due = now + self.long_click_min_duration
                    self.pressed = True
                else:
                    self._trigger(self.RELEASED)
                    if self.last_click != self.LONG:
                        self._trigger(self.last_click)
                    else:
                        self.last_click = self.SINGLE
                    self.pressed = False
            else:
                if self.pressed and self.click_enabled[self.LONG]:
                    if (
                        time.monotonic() > long_click_due
                        and self.last_click != self.LONG
                    ):
                        self.last_click = self.LONG
                        self._trigger(self.LONG)
            await asyncio.sleep(0)

    def _increase_clicks(self):
        if self.last_click == self.SINGLE and self.click_enabled[self.DOUBLE]:
            self.last_click = self.DOUBLE
        elif self.last_click == self.DOUBLE and self.click_enabled[self.TRIPLE]:
            self.last_click = self.TRIPLE
        else:
            self.last_click = self.SINGLE

    def _trigger(self, event: int):
        evt = self.events[event]
        evt.set()
        evt.clear()

    @staticmethod
    async def _set_event_when_done(coro: Awaitable, event: asyncio.Event):
        await coro
        event.set()

    async def wait(self, click_types: Sequence[int] = ALL_EVENTS):
        """
        Wait for the first of the specified events.

        :param List[int] click_types: One or more events to listen for. Default is to listen
          for all events
        :return: A list of the clicks that actually happened.

        :example:
          .. code-block:: python

            >>> async def get_click():
            >>>     # wait for a double or triple click
            >>>     clicks = await button.wait((Button.DOUBLE, Button.TRIPLE))
            >>>     if Button.DOUBLE in clicks:
            >>>         # do something

        """
        evts: Dict[int, asyncio.Task] = {}
        one_event_done = asyncio.Event()
        for evt_type in click_types:
            coro = self.events[evt_type].wait()
            evts[evt_type] = asyncio.create_task(
                self._set_event_when_done(coro, one_event_done)
            )
        await one_event_done.wait()
        if len(evts) > 1:
            await asyncio.sleep(0)  # ensure all event types get an opportunity to run
        results = []
        for evt_type, evt in evts.items():
            if evt.done():
                results.append(evt_type)
            else:
                evt.cancel()  # cancel unfired events
        return results

    async def wait_for_click(self):
        """
        Wait for any click and return it

        :return: Which click happened i.e. one of `SINGLE`, `DOUBLE`, `TRIPLE` or `LONG`
        """
        clicks = await self.wait(self.ANY_CLICK)
        return clicks[0]

    def deinit(self):
        """
        Deinitialise object and stop the background task
        """
        self.monitor_task.cancel()
        self.keys.deinit()



# SPDX-FileCopyrightText: Copyright (c) 2023 Phil Underwood for Underwood Underground
#
# SPDX-License-Identifier: MIT
from unittest import IsolatedAsyncioTestCase
from unittest.mock import patch, MagicMock
import sys
import asyncio

import microcontroller
import keypad

sys.modules["countio"] = MagicMock()

# pylint: disable=wrong-import-position
import async_button


class TestButton(IsolatedAsyncioTestCase):
    # pylint: disable=invalid-name, too-many-public-methods
    def setUp(self) -> None:
        self.time = MagicMock()
        self.patch1 = patch("async_button.time", new=self.time)
        self.patch1.start()
        self.time.monotonic.side_effect = self.new_monotonic
        self.keypad_keys = MagicMock()
        self.patch2 = patch("async_button.keypad.Keys", new=self.keypad_keys)
        self.patch2.start()
        self.keys = MagicMock()
        self.keypad_keys.return_value = self.keys
        self.keys.events.get_into = self.new_key_get
        self.button = None
        self.time_count = 0
        self.interval = 0.01
        self.button_timings = []
        self.button_state = False
        self.pin = microcontroller.Pin(0)

    async def asyncTearDown(self) -> None:
        self.patch1.stop()
        self.patch2.stop()
        if self.button:
            await asyncio.sleep(0)
            self.button.deinit()
            await asyncio.sleep(0)

    def new_monotonic(self) -> float:
        return self.time_count

    def new_key_get(self, event: keypad.Event) -> bool:
        self.time_count += self.interval
        if self.button_timings:
            if self.button_timings[0] <= self.time_count:
                self.button_state = not self.button_state
                self.button_timings.pop(0)
                # pylint: disable=protected-access
                event._pressed = self.button_state
                return True
        return False

    async def timeout(self):
        while True:
            if self.time_count > 10:
                raise TimeoutError
            await asyncio.sleep(0)

    async def wait_event_with_timeout(self, events):
        timeout = asyncio.create_task(self.timeout())
        button_wait = asyncio.create_task(self.button.wait(events))
        try:
            for coro in asyncio.as_completed((timeout, button_wait)):
                return await coro
        finally:
            timeout.cancel()
            button_wait.cancel()

    async def test_create_active_high(self):
        self.button = async_button.Button(self.pin, True)
        self.keypad_keys.assert_called_once_with(
            (self.pin,), value_when_pressed=True, pull=True
        )

    async def test_create_active_low(self):
        self.button = async_button.Button(self.pin, False)
        self.keypad_keys.assert_called_once_with(
            (self.pin,), value_when_pressed=False, pull=True
        )

    async def test_create_pull_false(self):
        self.button = async_button.Button(self.pin, True, pull=False)
        self.keypad_keys.assert_called_once_with(
            (self.pin,), value_when_pressed=True, pull=False
        )

    async def test_create_with_triple_but_no_double_fails(self):
        with self.assertRaises(ValueError):
            self.button = async_button.Button(
                self.pin, False, double_click_enable=False, triple_click_enable=True
            )

    async def test_button_pressed(self):
        self.button = async_button.Button(self.pin, True)
        self.button_timings = [0.10]
        await self.wait_event_with_timeout([async_button.Button.PRESSED])
        self.assertAlmostEqual(self.time_count, 0.10, delta=0.05)

    async def test_button_released(self):
        self.button = async_button.Button(self.pin, True)
        self.button_timings = [0.10, 0.20]
        await self.wait_event_with_timeout([async_button.Button.RELEASED])
        self.assertAlmostEqual(self.time_count, 0.20, delta=0.05)

    async def test_single_click(self):
        self.button = async_button.Button(self.pin, True)
        self.button_timings = [0.10, 0.20]
        await self.wait_event_with_timeout([async_button.Button.SINGLE])
        self.assertAlmostEqual(self.time_count, 0.20, delta=0.05)

    async def test_double_click(self):
        self.button = async_button.Button(self.pin, True)
        self.button_timings = [0.10, 0.30, 0.5, 0.7]
        await self.wait_event_with_timeout([async_button.Button.DOUBLE])
        self.assertAlmostEqual(self.time_count, 0.70, delta=0.05)

    async def test_double_click_not_when_disabled(self):
        self.button = async_button.Button(self.pin, True, double_click_enable=False)
        self.button_timings = [0.10, 0.30, 0.5, 0.7]
        with self.assertRaises(TimeoutError):
            await self.wait_event_with_timeout([async_button.Button.DOUBLE])

    async def test_double_click_not_when_too_slow(self):
        self.button = async_button.Button(
            self.pin,
            True,
        )
        self.button_timings = [0.10, 0.30, 1.1, 1.3]
        with self.assertRaises(TimeoutError):
            await self.wait_event_with_timeout([async_button.Button.DOUBLE])

    async def test_double_click_goes_back_to_single_after_gap(self):
        self.button = async_button.Button(self.pin, True)
        self.button_timings = [0.10, 0.30, 0.5, 0.7, 1.5, 1.7]
        await self.wait_event_with_timeout([async_button.Button.DOUBLE])
        self.assertAlmostEqual(self.time_count, 0.70, delta=0.05)
        await self.wait_event_with_timeout([async_button.Button.SINGLE])

    async def test_two_singles_when_too_slow(self):
        self.button = async_button.Button(
            self.pin,
            True,
        )
        self.button_timings = [0.10, 0.30, 1.1, 1.3]
        await self.wait_event_with_timeout([async_button.Button.SINGLE])
        await self.wait_event_with_timeout([async_button.Button.SINGLE])

    async def test_double_click_not_when_too_slow_with_different_time(self):
        self.button = async_button.Button(self.pin, True, double_click_max_duration=0.1)
        self.button_timings = [0.10, 0.30, 0.5, 0.7]
        with self.assertRaises(TimeoutError):
            await self.wait_event_with_timeout([async_button.Button.DOUBLE])

    async def test_triple_click(self):
        self.button = async_button.Button(self.pin, True, triple_click_enable=True)
        self.button_timings = [0.10, 0.30, 0.5, 0.7, 0.9, 1.1]
        await self.wait_event_with_timeout([async_button.Button.TRIPLE])
        self.assertAlmostEqual(self.time_count, 1.1, delta=0.05)

    async def test_triple_click_not_when_disabled(self):
        self.button = async_button.Button(self.pin, True)
        self.button_timings = [0.10, 0.30, 0.5, 0.7, 0.9, 1.1]
        with self.assertRaises(TimeoutError):
            await self.wait_event_with_timeout([async_button.Button.TRIPLE])

    async def test_long_click(self):
        self.button = async_button.Button(self.pin, True, long_click_enable=True)
        self.button_timings = [0.10, 3.30]
        await self.wait_event_with_timeout([async_button.Button.LONG])
        self.assertAlmostEqual(self.time_count, 2.1, delta=0.1)

    async def test_long_click_not_when_disabled(self):
        self.button = async_button.Button(self.pin, True, long_click_enable=False)
        self.button_timings = [0.10, 2.30]
        with self.assertRaises(TimeoutError):
            await self.wait_event_with_timeout([async_button.Button.LONG])

    async def test_long_click_does_not_produce_single_or_long_click_at_end(self):
        self.button = async_button.Button(self.pin, True, long_click_enable=True)
        self.button_timings = [0.10, 2.30]
        await self.wait_event_with_timeout([async_button.Button.LONG])
        with self.assertRaises(TimeoutError):
            await self.wait_event_with_timeout(
                [async_button.Button.SINGLE, async_button.Button.LONG]
            )

    async def test_single_then_double_then_triple_works(self):
        self.button = async_button.Button(self.pin, True, triple_click_enable=True)
        self.button_timings = [0.10, 0.30, 0.5, 0.7, 0.9, 1.1]
        await self.wait_event_with_timeout([async_button.Button.SINGLE])
        await self.wait_event_with_timeout([async_button.Button.DOUBLE])
        await self.wait_event_with_timeout([async_button.Button.TRIPLE])
        self.assertAlmostEqual(self.time_count, 1.10, delta=0.05)

    async def test_wait_for_clicks(self):
        self.button = async_button.Button(self.pin, True, triple_click_enable=True)
        self.button_timings = [0.10, 0.30, 0.5, 0.7, 0.9, 1.1]
        self.assertEqual(await self.button.wait_for_click(), self.button.SINGLE)
        self.assertEqual(await self.button.wait_for_click(), self.button.DOUBLE)
        self.assertEqual(await self.button.wait_for_click(), self.button.TRIPLE)
        self.assertAlmostEqual(self.time_count, 1.10, delta=0.05)

    async def test_wait_for_clicks_goes_back_to_single(self):
        self.button = async_button.Button(self.pin, True)
        self.button_timings = [0.10, 0.30, 0.5, 0.7, 0.9, 1.1]
        self.assertEqual(await self.button.wait_for_click(), self.button.SINGLE)
        self.assertEqual(await self.button.wait_for_click(), self.button.DOUBLE)
        self.assertEqual(await self.button.wait_for_click(), self.button.SINGLE)
        self.assertAlmostEqual(self.time_count, 1.10, delta=0.05)

    async def test_wait_selection(self):
        self.button = async_button.Button(self.pin, True)
        selection = (self.button.SINGLE, self.button.PRESSED)
        self.button_timings = [0.10, 0.30, 0.5, 0.7, 0.9, 1.1]
        self.assertSequenceEqual(
            await self.button.wait(selection), (self.button.PRESSED,)
        )
        self.assertSequenceEqual(
            await self.button.wait(selection), (self.button.SINGLE,)
        )
        self.assertSequenceEqual(
            await self.button.wait(selection), (self.button.PRESSED,)
        )
        self.assertSequenceEqual(
            await self.button.wait(selection), (self.button.PRESSED,)
        )
        self.assertSequenceEqual(
            await self.button.wait(selection), (self.button.SINGLE,)
        )
        self.assertAlmostEqual(self.time_count, 1.10, delta=0.05)

    async def test_wait_all(self):
        self.button = async_button.Button(self.pin, True, triple_click_enable=True)
        self.button_timings = [0.10, 0.30, 0.5, 0.7, 0.9, 1.1]
        self.assertSequenceEqual(await self.button.wait(), (self.button.PRESSED,))
        self.assertSequenceEqual(
            await self.button.wait(), (self.button.RELEASED, self.button.SINGLE)
        )
        self.assertSequenceEqual(await self.button.wait(), (self.button.PRESSED,))
        self.assertSequenceEqual(
            await self.button.wait(), (self.button.RELEASED, self.button.DOUBLE)
        )
        self.assertSequenceEqual(await self.button.wait(), (self.button.PRESSED,))
        self.assertSequenceEqual(
            await self.button.wait(), (self.button.RELEASED, self.button.TRIPLE)
        )
        self.assertAlmostEqual(self.time_count, 1.10, delta=0.05)



# SPDX-FileCopyrightText: 2017 Scott Shawcroft, written for Adafruit Industries
# SPDX-FileCopyrightText: Copyright (c) 2023 Phil Underwood for Underwood Underground
#
# SPDX-License-Identifier: Unlicense
import asyncio

import board
import digitalio

from microcontroller import Pin

from async_button import Button


def setUpLed(pin: Pin):
    dio = digitalio.DigitalInOut(pin)
    dio.direction = digitalio.Direction.OUTPUT
    dio.value = True
    return dio


CLICK_NAMES = {
    Button.SINGLE: "Single click",
    Button.DOUBLE: "Double click",
    Button.TRIPLE: "Triple click",
    Button.LONG: "Long click",
}


async def counter():
    i = 0
    while True:
        print(f"COUNTER: {i}")
        await asyncio.sleep(1)
        i += 1


async def button_led_watcher(button: Button, led: digitalio.DigitalInOut, click):
    while True:
        await button.wait((click,))
        led.value = False
        await asyncio.sleep(0.5)
        led.value = True


async def click_watcher(button: Button):
    while True:
        click = await button.wait_for_click()
        print(f"{CLICK_NAMES[click]} seen")


async def main():
    # note Button must be created in an async environment
    button = Button(
        board.D5,
        value_when_pressed=False,
        triple_click_enable=True,
        long_click_enable=True,
    )
    green = setUpLed(board.D9)
    red = setUpLed(board.D8)
    blue = setUpLed(board.D7)

    red_watcher = button_led_watcher(button, red, Button.SINGLE)
    green_watcher = button_led_watcher(button, green, Button.DOUBLE)
    blue_watcher = button_led_watcher(button, blue, Button.TRIPLE)
    await asyncio.gather(
        counter(), click_watcher(button), red_watcher, green_watcher, blue_watcher
    )


asyncio.run(main())



# SPDX-FileCopyrightText: 2017 Scott Shawcroft, written for Adafruit Industries
# SPDX-FileCopyrightText: Copyright (c) 2023 Phil Underwood for Underwood Underground
#
# SPDX-License-Identifier: Unlicense
import asyncio

import board
import digitalio

import async_button

# green = digitalio.DigitalInOut(board.D9)
# red = digitalio.DigitalInOut(board.D8)
# blue = digitalio.DigitalInOut(board.D7)


async def counter():
    i = 0
    while True:
        print(f"COUNTER: {i}")
        await asyncio.sleep(1)
        i += 1


async def button_watcher():
    button = async_button.SimpleButton(board.D5, value_when_pressed=False)
    led = digitalio.DigitalInOut(board.D8)
    led.direction = digitalio.Direction.OUTPUT
    led.value = True
    while True:
        await button.pressed()
        print("Button pressed")
        led.value = 0
        await button.released()
        print("Button released")
        led.value = 1


async def main():
    await asyncio.gather(counter(), button_watcher())


asyncio.run(main())



# SPDX-FileCopyrightText: 2017 Scott Shawcroft, written for Adafruit Industries
# SPDX-FileCopyrightText: Copyright (c) 2023 Phil Underwood for Underwood Underground
#
# SPDX-License-Identifier: MIT
"""
Tests for Simple Button implementation
"""

from unittest import IsolatedAsyncioTestCase
from unittest.mock import patch, MagicMock, PropertyMock, AsyncMock, Mock

import digitalio


class TestSimpleButton(IsolatedAsyncioTestCase):
    def setUp(self) -> None:
        self.countio = MagicMock()
        self.counter = MagicMock()
        self.counter.__enter__.return_value = (
            self.counter
        )  # return self as a context manager
        type(self.counter).count = PropertyMock(side_effect=[0, 0, 1])
        self.countio.Counter.return_value = self.counter
        self.asyncio = Mock()
        self.asyncio.sleep = AsyncMock()
        self.edge_rise = self.countio.Edge.RISE
        self.edge_fall = self.countio.Edge.FALL
        with patch.dict("sys.modules", countio=self.countio, asyncio=self.asyncio):
            # pylint: disable=import-outside-toplevel
            from async_button import SimpleButton

            self.simple_button_class = SimpleButton

    async def test_pressed_active_high(self):
        button = self.simple_button_class("P1", True)
        await button.pressed()
        self.countio.Counter.assert_called_once_with(
            "P1", self.edge_rise, digitalio.Pull.DOWN
        )
        self.assertEqual(self.asyncio.sleep.await_count, 2)

    async def test_released_active_high(self):
        button = self.simple_button_class("P1", True)
        await button.released()
        self.countio.Counter.assert_called_once_with(
            "P1", self.edge_fall, digitalio.Pull.DOWN
        )
        self.assertEqual(self.asyncio.sleep.await_count, 2)

    async def test_pressed_active_low(self):
        button = self.simple_button_class("P1", False)
        await button.pressed()
        self.countio.Counter.assert_called_once_with(
            "P1", self.edge_fall, digitalio.Pull.UP
        )
        self.assertEqual(self.asyncio.sleep.await_count, 2)

    async def test_released_active_low(self):
        button = self.simple_button_class("P1", False)
        await button.released()
        self.countio.Counter.assert_called_once_with(
            "P1", self.edge_rise, digitalio.Pull.UP
        )
        self.assertEqual(self.asyncio.sleep.await_count, 2)
